// CICSClient.cpp : Implementation file for MQSI V2.0.1 plug-in node
//
// Template built by MQSI V2 Visual Studio Wizard (kolban@us.ibm.com)
//
// Installation:
// For NT:		<MQSI Root>\bin
// For Unix:	<MQSI Root>/lil
//
#ifdef _WIN32
#include "stdafx.h"
#endif

#include <stdarg.h>		// Required for variable length arguments
#include <stdio.h>		// Required for file IO and NULL declaration
#include <stdlib.h>		// Required for memory manipulation
#include <BipCci.h>		// Common functions (between nodes and parsers)
#include <BipCni.h>		// Message processing nodes
#include <CICS_ECI.H>	// Include the CICS Universal Client header
#ifdef _WIN32
#include "NT_msg/cicsclient_msg.h"
#endif

#include "NodeUtils.h"	// Include the required utility functions



///////////////////////////////////////////////////////////
// NodeContext
struct NodeContext
{
	MQSIX_ATTRIBUTES	*attributes;
	CciTerminal	*inTerminal;		// In Terminal
	CciTerminal	*outTerminal;		// Out Terminal
	CciTerminal	*failureTerminal;	// Out Terminal
	char	programName[9];
	char	userid[9];
	char	password[9];
	char	region[9];
	int		commareaLength;
	int		removeMQMD;
};

#ifdef _WIN32
extern "C"
{
	void __declspec(dllexport) *bipGetMessageflowNodeFactory();
};

HMODULE	g_hModule;

BOOL APIENTRY DllMain(
	HMODULE hModule,
	DWORD  ul_reason_for_call,
	LPVOID lpReserved
					 )
{
	g_hModule = hModule;
    return TRUE;
}
#endif


///////////////////////////////////////////////////////////
// CreateNodeContext
void *CreateNodeContext(
	CciFactory	*factoryObject,
	CciChar		*nodeName,
	CciNode		*nodeObject)
{
	struct NodeContext	*pMyContext;
	int					iRc;

	pMyContext = (struct NodeContext *)malloc(sizeof(struct NodeContext));

	mqsixLog(">> CreateNodeContext: Node Name=%S\n", nodeName);
	pMyContext->inTerminal = cniCreateInputTerminal(&iRc, nodeObject, L"in");
	pMyContext->outTerminal = cniCreateOutputTerminal(&iRc, nodeObject, L"out");
	pMyContext->failureTerminal = cniCreateOutputTerminal(&iRc, nodeObject, L"failure");
	pMyContext->attributes = NULL;
	mqsixAddAttribute(&pMyContext->attributes, L"regionName", L"");
	mqsixAddAttribute(&pMyContext->attributes, L"programName", L"");
	mqsixAddAttribute(&pMyContext->attributes, L"userid", L"");
	mqsixAddAttribute(&pMyContext->attributes, L"password", L"");
	mqsixAddAttribute(&pMyContext->attributes, L"commareaLength", L"");
	mqsixAddAttribute(&pMyContext->attributes, L"removeMQMD", L"");

	memset(pMyContext->region, 0, 8);
	memset(pMyContext->programName, 0, 8);
	memset(pMyContext->userid, 0, 8);
	memset(pMyContext->password, 0, 8);
	pMyContext->commareaLength = 0;
	pMyContext->removeMQMD = 1;

	return(pMyContext);
}

///////////////////////////////////////////////////////////
// DeleteNodeContext
void DeleteNodeContext(CciContext *context)
{
	struct NodeContext	*pMyContext;

	pMyContext = (struct NodeContext *)context;
	mqsixLog(">> DeleteNodeContext\n");
	mqsixReleaseAttributes(pMyContext->attributes);
	pMyContext->attributes = NULL;
	free(context);
}

///////////////////////////////////////////////////////////
// GetAttributeName
int GetAttributeName(
	CciContext	*context,
	int			index,
	CciChar		*buffer,
	int			bufsize)
{
	int	iRc;
	struct NodeContext	*pMyContext;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixGetAttributeName(
		pMyContext->attributes,
		index,
		buffer,
		bufsize);
	return(iRc);
}

///////////////////////////////////////////////////////////
// GetAttribute
int GetAttribute(
	CciContext	*context,
	CciChar		*attrName,
	CciChar		*buffer,
	int bufsize)
{
	struct NodeContext	*pMyContext;
	int					iRc;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixGetAttribute(
		pMyContext->attributes,
		attrName,
		buffer,
		bufsize);
	return(iRc);
}

///////////////////////////////////////////////////////////
// SetAttribute
int SetAttribute(
	CciContext	*context,
	CciChar		*attrName,
	CciChar		*attrValue)
{
	struct NodeContext	*pMyContext;
	int					iRc;
	char				*pValue;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixSetAttribute(pMyContext->attributes,
		attrName,
		attrValue);


	// Region Name
	if (mqsixStringCompare(attrName, L"regionName") == 0)
	{
		pValue = mqsixUnicodeToASCII(attrValue);
		sprintf(pMyContext->region, "%-8.8s", pValue);
		free(pValue);
	}
	// Program Name
	if (mqsixStringCompare(attrName, L"programName") == 0)
	{
		pValue = mqsixUnicodeToASCII(attrValue);
		sprintf(pMyContext->programName, "%-8.8s", pValue);
		free(pValue);
	}
	// Userid
	if (mqsixStringCompare(attrName, L"userid") == 0)
	{
		pValue = mqsixUnicodeToASCII(attrValue);
		sprintf(pMyContext->userid, "%-8.8s", pValue);
		free(pValue);
	}
	// Password
	if (mqsixStringCompare(attrName, L"password") == 0)
	{
		pValue = mqsixUnicodeToASCII(attrValue);
		sprintf(pMyContext->password, "%-8.8s", pValue);
		free(pValue);
	}
	// Commarea Length
	if (mqsixStringCompare(attrName, L"commareaLength") == 0)
	{
		pValue = mqsixUnicodeToASCII(attrValue);
		pMyContext->commareaLength = atoi(pValue);
		free(pValue);
	}
	// Remove MQMD
	if (mqsixStringCompare(attrName, L"removeMQMD") == 0)
	{
		pValue = mqsixUnicodeToASCII(attrValue);
		if (mqsixStringCompare(attrValue, L"true") == 0)
		{
			pMyContext->removeMQMD = 1;
		}
		else
		{
			pMyContext->removeMQMD = 0;
		}
		free(pValue);
	}
	return(iRc);
}

///////////////////////////////////////////////////////////
// CICSErrorToString
//
char *CICSErrorToString(int rc)
{
	switch(rc)
	{
	case ECI_NO_ERROR:
		return("ECI_NO_ERROR");
	case ECI_ERR_INVALID_DATA_LENGTH:
		return("ECI_ERR_INVALID_DATA_LENGTH");
	case ECI_ERR_INVALID_EXTEND_MODE:
		return("ECI_ERR_INVALID_EXTEND_MODE");
	case ECI_ERR_NO_CICS:
		return("ECI_ERR_NO_CICS");
	case ECI_ERR_CICS_DIED:
		return("ECI_ERR_CICS_DIED");
	case ECI_ERR_REQUEST_TIMEOUT:
		return("ECI_ERR_REQUEST_TIMEOUT");
	/*
	case ECI_ERR_NO_REPLY:
		return("ECI_ERR_NO_REPLY");
	*/
	case ECI_ERR_RESPONSE_TIMEOUT:
		return("ECI_ERR_RESPONSE_TIMEOUT");
	case ECI_ERR_TRANSACTION_ABEND:
		return("ECI_ERR_TRANSACTION_ABEND");
	case ECI_ERR_EXEC_NOT_RESIDENT:
		return("ECI_ERR_EXEC_NOT_RESIDENT");
	case ECI_ERR_SYSTEM_ERROR:
		return("ECI_ERR_SYSTEM_ERROR");
	case ECI_ERR_NULL_WIN_HANDLE:
		return("ECI_ERR_NULL_WIN_HANDLE");
	case ECI_ERR_NULL_MESSAGE_ID:
		return("ECI_ERR_NULL_MESSAGE_ID");
	case ECI_ERR_THREAD_CREATE_ERROR:
		return("ECI_ERR_THREAD_CREATE_ERROR");
	case ECI_ERR_INVALID_CALL_TYPE:
		return("ECI_ERR_INVALID_CALL_TYPE");
	case ECI_ERR_ALREADY_ACTIVE:
		return("ECI_ERR_ALREADY_ACTIVE");
	case ECI_ERR_RESOURCE_SHORTAGE:
		return("ECI_ERR_RESOURCE_SHORTAGE");
	case ECI_ERR_NO_SESSIONS:       
		return("ECI_ERR_NO_SESSIONS");
	case ECI_ERR_NULL_SEM_HANDLE:
		return("ECI_ERR_NULL_SEM_HANDLE");
	case ECI_ERR_INVALID_DATA_AREA:
		return("ECI_ERR_INVALID_DATA_AREA");
	case ECI_ERR_INVALID_VERSION:
		return("ECI_ERR_INVALID_VERSION");
	case ECI_ERR_UNKNOWN_SERVER:
		return("ECI_ERR_UNKNOWN_SERVER");
	case ECI_ERR_CALL_FROM_CALLBACK:
		return("ECI_ERR_CALL_FROM_CALLBACK");
	case ECI_ERR_INVALID_TRANSID:
		return("ECI_ERR_INVALID_TRANSID");
	case ECI_ERR_MORE_SYSTEMS:
		return("ECI_ERR_MORE_SYSTEMS");
	case ECI_ERR_NO_SYSTEMS:
		return("ECI_ERR_NO_SYSTEMS");
	case ECI_ERR_SECURITY_ERROR:
		return("ECI_ERR_SECURITY_ERROR");
	case ECI_ERR_MAX_SYSTEMS:
		return("ECI_ERR_MAX_SYSTEMS");
	case ECI_ERR_MAX_SESSIONS:
		return("ECI_ERR_MAX_SESSIONS");
	case ECI_ERR_ROLLEDBACK:
		return("ECI_ERR_ROLLEDBACK");
	default:
		return("*** Unknown ***");
	}
}


///////////////////////////////////////////////////////////
// Call CICS
int CallCICS(char *regionName,
			  char *programName,
			  char *userid,
			  char *password,
			  char *commarea,
			  int commareaLength)
{
	
	int			cicsRc;
	int			iRc;
	ECI_PARMS	parms;
	char		messageText[100];


	mqsixLog("About to call CICS:\n");
	mqsixLog("\tRegion: \"%.8s\"\n", regionName);
	mqsixLog("\tProgram: \"%.8s\"\n", programName);
	mqsixLog("\tUserid: \"%.8s\"\n", userid);
	mqsixLog("\tPassword: \"%.8s\"\n", password);
	mqsixLog("\tCommarea length: %d\n", commareaLength);

	memset(&parms, 0, sizeof(parms));
	parms.eci_version = ECI_VERSION_1A;
	strncpy(parms.eci_system_name, regionName, 8);
	parms.eci_call_type = ECI_SYNC;
	strncpy(parms.eci_program_name, programName, 8);
	if (commareaLength == 0)
	{
		parms.eci_commarea = NULL;
	}
	else
	{
		parms.eci_commarea = commarea;
	}
	parms.eci_commarea_length = commareaLength;
	memcpy(parms.eci_userid, userid, 8);
	memcpy(parms.eci_password, password, 8);

	cicsRc = CICS_ExternalCall(&parms);
	mqsixLog("Return code from ECI Call: %d\n", cicsRc);
	if (cicsRc != 0)
	{
		if (cicsRc == ECI_ERR_TRANSACTION_ABEND)
		{
			sprintf(messageText, "%d Abend Code: '%.4s'", cicsRc, parms.eci_abend_code);
		}
		else
		{
			sprintf(messageText, "%d", cicsRc);
		}
		cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "Evaluate",
			L"MQSIV2_CICSCLIENT", MSG_CICS_ECI_ERROR,
			"Error response from CICS Client: %1 %2",
			messageText, CICSErrorToString(cicsRc), (char *)0);
	}
	return(cicsRc);
}

///////////////////////////////////////////////////////////
// BuildBlobMessage
CciMessage *BuildBlobMessage(CciMessage *originalMessage, CciByte *pData, CciSize iSize)
{
	CciMessage	*tempMessage;
	CciElement	*pElement;
	CciElement	*pElement2;
	CciElement	*pRoot;
	int			iRc;
	CciByteArray	byteArray;
	tempMessage = cniCreateMessage(&iRc, cniGetMessageContext(&iRc, originalMessage));
	pRoot = cniRootElement(&iRc, tempMessage);
	pElement = cniSearchFirstChild(&iRc,
		cniRootElement(&iRc, originalMessage),
		CCI_COMPARE_MODE_NAME,
		0,
		L"Properties");
	if (pElement != NULL)
	{
		// Add a child called "Properties"
		pElement2 = cniCreateElementAsLastChild(&iRc, pRoot);
		cniSetElementName(&iRc, pElement2, L"Properties");
		cniSetElementType(&iRc, pElement2, CCI_ELEMENT_TYPE_NAME);

		// Copy the tree ...
		cniCopyElementTree(&iRc, pElement, pElement2);
	}
	pElement = cniSearchFirstChild(&iRc, cniRootElement(&iRc, originalMessage), CCI_COMPARE_MODE_NAME,
			0, L"MQMD");
	if (pElement != NULL)
	{
		// Add a child called "MQMD"
		pElement2 = cniCreateElementAsLastChildUsingParser(&iRc, pRoot, L"MQHMD");
		if (iRc == 64)
		{
			cciRethrowLastException(&iRc);
		}
		cniSetElementName(&iRc, pElement2, L"MQMD");
		cniSetElementType(&iRc, pElement2, CCI_ELEMENT_TYPE_NAME);

		// Copy the tree ...
		cniCopyElementTree(&iRc, pElement, pElement2);
	}
	pElement = cniCreateElementAsLastChildUsingParser(&iRc, pRoot, L"NONE");
	cniSetElementName(&iRc, pElement, L"BLOB");
	cniSetElementType(&iRc, pElement, CCI_ELEMENT_TYPE_NAME);

	pElement2 = cniCreateElementAsLastChild(&iRc, pElement);
	cniSetElementName(&iRc, pElement2, L"UnknownParserName");
	cniSetElementType(&iRc, pElement2, CCI_ELEMENT_TYPE_NAME_VALUE);
	cniSetElementCharacterValue(&iRc, pElement2, L"MQSTR", 5);
	
	pElement = cniCreateElementAsLastChild(&iRc, pElement);
	cniSetElementName(&iRc, pElement, L"BLOB");
	cniSetElementType(&iRc, pElement, CCI_ELEMENT_TYPE_NAME_VALUE);
	byteArray.pointer = pData;
	byteArray.size = iSize;
	cniSetElementByteArrayValue(&iRc, pElement, &byteArray);
	
	cniWriteBuffer(&iRc, tempMessage);
	iSize = cniBufferSize(&iRc, tempMessage);
	return(tempMessage);
}


///////////////////////////////////////////////////////////
// Evaluate
void Evaluate(
	CciContext	*context, 
	CciMessage	*inDestinationList,
	CciMessage	*inExceptionList,
	CciMessage	*inMessage)
{
	int					iRc;
	int					iRc2;
	struct NodeContext	*pMyContext;
	CciSize				iSize;
	const CciByte		*pData;
	CciByte				*pTempData = NULL;
	CciMessage			*tempMessage;
	CciMessage			*outMessage;
	CciElement			*pElement;
	int					iLocalCommareaLength;
	int					cicsRc;
	mqsixDebugBreak();

	mqsixLog(">> Evaluate\n");
	pMyContext = (struct NodeContext *)context;

	//
	// If the node attriutes ask for the removal of the MQMD, we remove at this
	// point.
	//
	if (pMyContext->removeMQMD)
	{
		mqsixLog("\tRemoving MQMD ...\n");
		tempMessage = cniCreateMessage(&iRc, cniGetMessageContext(&iRc, inMessage));
		cniCopyElementTree(&iRc, cniRootElement(&iRc, inMessage), cniRootElement(&iRc, tempMessage));
		pElement = cniSearchFirstChild(&iRc, cniRootElement(&iRc, tempMessage), CCI_COMPARE_MODE_NAME,
			0, L"MQMD");
		if (pElement != 0)
		{
			cniDetach(&iRc, pElement);
		}
	}
	else
	{
		tempMessage = inMessage;
	}

	//
	// Generate binary stream for Commarea ...
	// We do this by asking MQSI to build a message buffer from the current
	// message tree.  Next we look and see what (if any) commarea length has been
	// supplied by user on the node attributes.  If a length has been supplied,
	// we will use that value.  We will copy the bytearray message into this
	// commeara length storage, truncating as needed.
	//
	cniWriteBuffer(&iRc, tempMessage);
	iSize = cniBufferSize(&iRc, tempMessage);
	pData = cniBufferPointer(&iRc, tempMessage);

	if (pMyContext->commareaLength > 0)
	{
		iLocalCommareaLength = pMyContext->commareaLength;
		pTempData = (CciByte *)malloc(iLocalCommareaLength);
		memset(pTempData, 0, iLocalCommareaLength);
		if (iLocalCommareaLength > iSize)
		{
			memcpy(pTempData, pData, iSize);
		}
		else
		{
			memcpy(pTempData, pData, iLocalCommareaLength);
		}
	}
	else
	{
		pTempData = (CciByte *)malloc(iSize);
		memcpy(pTempData, pData, iSize);
		iLocalCommareaLength = iSize;
	}

	// Now call CICS!!
	cicsRc = CallCICS(pMyContext->region,
		pMyContext->programName,
		pMyContext->userid,
		pMyContext->password,
		(char *)pTempData, iLocalCommareaLength);
	if (tempMessage != inMessage)
	{
		cniDeleteMessage(&iRc, tempMessage);
	}

	outMessage = BuildBlobMessage(inMessage, (char *)pTempData, iLocalCommareaLength);
	if (pTempData != NULL)
	{
		free(pTempData);
	}

	// Propagate the message based upon the results from CICS call....
	if (cicsRc == 0)
	{
		cniPropagate(&iRc, pMyContext->outTerminal, 
			inDestinationList, inExceptionList, outMessage);
		mqsixLog("-- Propagated through out terminal...\n");

		cniDeleteMessage(&iRc2, outMessage);
	}
	else
	{
		cniDeleteMessage(&iRc2, outMessage);
		cniPropagate(&iRc, pMyContext->failureTerminal, 
			inDestinationList, inExceptionList, inMessage);
		mqsixLog("-- Propagated through failure terminal...\n");
	}
	
	if (iRc != 0)
	{
		mqsixLog("An exception was reported on return from propogate through 'out': %d\n", iRc);
		cciRethrowLastException(&iRc);
		return;
	}
	mqsixLog("<< Evaluate\n");
	return;
}

///////////////////////////////////////////////////////////
// bipGetMessageflowNodeFactory
// 
// Entry point into the CICSClient node.  Called by the broker when
// the node is initially loaded.


void __declspec(dllexport) *bipGetMessageflowNodeFactory()
{
	static CNI_VFT	virtualFunctionTable;
	CciChar			factoryName[] = L"CICSClientFactory";
	int				iRc;
	void			*factoryObject;

	mqsixRegisterMessages(g_hModule, "MQSIV2_CICSCLIENT");
	mqsixLog("\n");
	mqsixLog("************************\n");
	mqsixLog("*** Loading Node LIL ***\n");
	mqsixLog("************************\n");
	mqsixLog(">> Creating nodeFactory: %S, lil build: %s, CCI Version: %d.%d\n",
		factoryName, __TIMESTAMP__, CCI_VERSION >> 16, CCI_VERSION & 0xFFFF);
	memset(&virtualFunctionTable, 0, sizeof(virtualFunctionTable));
	virtualFunctionTable.iFpCreateNodeContext	= CreateNodeContext;
	virtualFunctionTable.iFpDeleteNodeContext	= DeleteNodeContext;
	virtualFunctionTable.iFpGetAttributeName	= GetAttributeName;
	virtualFunctionTable.iFpSetAttribute		= SetAttribute;
	virtualFunctionTable.iFpGetAttribute		= GetAttribute;
	virtualFunctionTable.iFpEvaluate			= Evaluate;

	factoryObject = cniCreateNodeFactory(&iRc, factoryName);
	cniDefineNodeClass(&iRc, factoryObject, L"CICSClientNode", &virtualFunctionTable); 

	return(factoryObject);
}
